import os
import smbc
import settings
import nose
from nose.plugins.skip import SkipTest


""" Test acl parser


"""

class SmbAcl():
    revision = None
    owner = None
    group = None
    acl = []

    def __init__(self, xattr_s=None):
        """ parse an acl into a SmbAcl object """
        if xattr_s:
            xattr_l = [ s[s.index(":")+1:] for s in xattr_s.split(",")]
            (self.revision, self.owner, self.group) = xattr_l[0:3]
            self.acl = xattr_l[3:] 


    def __str__(self):
        ret = "REVISION:%s,OWNER:%s,GROUP:%s" % (self.revision,self.owner,self.group)
        for a in self.acl:
            ret = ret + ",ACL:%s" % a
        return ret

    @staticmethod
    def get_target(acl_s):
        return acl_s[0:acl_s.index(":")]

    @staticmethod
    def get_perm(acl_s):
        return acl_s[acl_s.index(":")+1:]


"""SmbAcl parser To be implemented"""
def test_acl_parser():
    raise SkipTest("SmbAcl to be implemented")

    xattr_l = ["REVISION:1,OWNER:S-1-5-21-833659924-920326847-3160110649-3002,GROUP:S-1-22-2-1002,ACL:S-1-5-21-833659924-920326847-3160110649-3002:0/0/0x001e01ff,ACL:S-1-22-2-1002:0/0/0x00120089,ACL:S-1-1-0:0/0/0x00120089",
   #attrs(system.nt_sec_desc.*+):
   'REVISION:1,OWNER:RPOLLI\\babel,GROUP:Unix Group\\babel,ACL:RPOLLI\\babel:0/0/0x001e01ff,ACL:Unix Group\\babel:0/0/0x00120089,ACL:\Everyone:0/0/0x00120089']
    
    for xattr_s in xattr_l: 
        xattr = SmbAcl(xattr_s)
        print("xattr: %s" % xattr)
        
        assert xattr.revision
        assert xattr.owner
        assert xattr.group
        assert xattr.acl and len(xattr.acl)>1

def test_acl_get_target():
    acl_l = ["S-1-5-21-833659924-920326847-3160110649-3002:0/0/0x001e01ff", 
	"Unix Group\\babel:0/0/0x00120089"]
    for acl_s in acl_l:
        perm = SmbAcl.get_perm(acl_s)
        target = SmbAcl.get_target(acl_s)
      
        assert target 
        assert perm
     



#!/usr/bin/env python

import smbc
import settings
import hashlib
import os

basedir = 'smb://' + settings.SERVER + '/' + settings.SHARE + '/'
testdir = basedir  + '/' + settings.TESTDIR

def auth_fn(server, share, workgroup, username, password):
    return (workgroup, settings.USERNAME, settings.PASSWORD)

def creat_randfile(path, size):
    rand = open('/dev/urandom', 'rb')
    file = open(path, 'wb')
    buf = rand.read(size)
    file.write(buf)
    file.close()
    rand.close()

def md(path):
    m = hashlib.md5()
    file = open(path, 'rb')
    m.update(file.read())
    file.close()
    return m.hexdigest()

def upload_simple(spath, ctx, duri):
    sfile = open(spath, 'rb')
    dfile = ctx.open(duri, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
    ret = dfile.write(sfile.read())
    if ret < 0:
        raise IOError("smbc write error")
    sfile.close()
    dfile.close()
    return True

def download_simple(ctx, suri, dpath):
    sfile = ctx.open(suri, os.O_RDONLY)
    dfile = open(dpath, 'wb')
    dfile.write(sfile.read())
    dfile.flush()
    sfile.close()
    dfile.close()
    return True

def upload_buffer(spath, ctx, duri):
    sfile = open(spath, 'rb')
    dfile = ctx.open(duri, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
    while True:
        buf = sfile.read(8192)
        if not buf:
            break
        ret = dfile.write(buf)
        if ret < 0:
            raise IOError("smbc write error")
    sfile.close()
    dfile.close()
    return True

def download_buffer(ctx, suri, dpath):
    sfile = ctx.open(suri, os.O_RDONLY)
    dfile = open(dpath, 'wb')
    while True:
        buf = sfile.read(8192)
        if not buf:
            break
        dfile.write(buf)
    dfile.flush()
    sfile.close()
    dfile.close()
    return True

def upload_iter(spath, ctx, duri):
    sfile = open(spath, 'rb')
    dfile = ctx.open(duri, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
    for buf in sfile:
        ret = dfile.write(buf)
        if ret < 0:
            raise IOError("smbc write error")
    sfile.close()
    dfile.close()
    return True

def download_iter(ctx, suri, dpath):
    sfile = ctx.open(suri, os.O_RDONLY)
    dfile = open(dpath, 'wb')
    for buf in sfile:
        dfile.write(buf)
    dfile.flush()
    sfile.close()
    dfile.close()
    return True

def setUp():
    global ctx
    ctx = smbc.Context()
    ctx.optionNoAutoAnonymousLogin = True
    ctx.functionAuthData = auth_fn

def tearDown():
    global ctx
    del ctx

def test_copy_simple_1k():
    creat_randfile('test1.dat', 1000)
    testfile = basedir + '/' + 'test1.dat'
    upload_simple('test1.dat', ctx, testfile)
    download_simple(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))

def test_copy_simple_100k():
    creat_randfile('test1.dat', 10000)
    testfile = basedir + '/' + 'test1.dat'
    upload_simple('test1.dat', ctx, testfile)
    download_simple(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))

def test_copy_simple_1M():
    creat_randfile('test1.dat', 1000000)
    testfile = basedir + '/' + 'test1.dat'
    upload_simple('test1.dat', ctx, testfile)
    download_simple(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))

def test_copy_buffer_1k():
    creat_randfile('test1.dat', 1000)
    testfile = basedir + '/' + 'test1.dat'
    upload_buffer('test1.dat', ctx, testfile)
    download_buffer(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))

def test_copy_buffer_100k():
    creat_randfile('test1.dat', 10000)
    testfile = basedir + '/' + 'test1.dat'
    upload_buffer('test1.dat', ctx, testfile)
    download_buffer(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))

def test_copy_buffer_1M():
    creat_randfile('test1.dat', 1000000)
    testfile = basedir + '/' + 'test1.dat'
    upload_buffer('test1.dat', ctx, testfile)
    download_buffer(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))

def test_copy_iter_1k():
    creat_randfile('test1.dat', 1000)
    testfile = basedir + '/' + 'test1.dat'
    upload_iter('test1.dat', ctx, testfile)
    download_iter(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))

def test_copy_iter_100k():
    creat_randfile('test1.dat', 10000)
    testfile = basedir + '/' + 'test1.dat'
    upload_iter('test1.dat', ctx, testfile)
    download_iter(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))

def test_copy_iter_1M():
    creat_randfile('test1.dat', 1000000)
    testfile = basedir + '/' + 'test1.dat'
    upload_iter('test1.dat', ctx, testfile)
    download_iter(ctx, testfile, 'test2.dat')
    assert(md('test1.dat') == md('test2.dat'))



#!/usr/bin/python


SMB_POSIX_ACL_USER_OBJ=0x01
SMB_POSIX_ACL_USER=0x02
SMB_POSIX_ACL_GROUP_OBJ=0x04
SMB_POSIX_ACL_GROUP=0x08
SMB_POSIX_ACL_MASK=0x10
SMB_POSIX_ACL_OTHER=0x20


SMB_POSIX_ACL_READ=0x04
SMB_POSIX_ACL_WRITE=0x02
SMB_POSIX_ACL_EXECUTE=0x01

SMB_POSIX_ACL_HEADER_SIZE=6
SMB_POSIX_ACL_ENTRY_SIZE=10


rwx=0x001e01ff
x=0x001200a0
w=0x00120116
r=0x00120089

class SmbAcl():
    revision = None
    owner = None
    group = None
    acl = []

    def __init__(self, xattr_s=None):
        """ parse an acl into a SmbAcl object """
        if xattr_s:
            xattr_l = [ s[s.index(":")+1:] for s in xattr_s.split(",")]
            (self.revision, self.owner, self.group) = xattr_l[0:3]
            self.acl = xattr_l[3:]


    def __str__(self):
        ret = "REVISION:%s,OWNER:%s,GROUP:%s" % (self.revision,self.owner,self.group)
        for a in self.acl:
            ret = ret + ",ACL:%s" % a
        return ret

    @staticmethod
    def get_target(acl_s):
        return acl_s[0:acl_s.index(":")]

    @staticmethod
    def get_perm(acl_s):
        return acl_s[acl_s.index(":")+1:]

#
#import smbc
#
#c = smbc.Context(debug=1)
#cb = lambda se, sh, w, u, p: ("WORKGROUP","babel","b3nedetto")
#c.functionAuthData = cb
#
#print c.opendir("smb://localhost/share/Neffa").getdents()
#print c.getxattr("smb://localhost/share/Neffa", "system.nt_sec_desc.*")
#print c.getxattr("smb://localhost/share/Neffa", "s")
#



#!/usr/bin/env python

import smbc
import settings
import stat

basedir = 'smb://' + settings.SERVER + '/' + settings.SHARE + '/'
testdir = basedir  + '/' + settings.TESTDIR

def auth_fn(server, share, workgroup, username, password):
    return (workgroup, settings.USERNAME, settings.PASSWORD)

def setUp():
    global ctx
    ctx = smbc.Context()
    ctx.optionNoAutoAnonymousLogin = True
    ctx.functionAuthData = auth_fn

def tearDown():
    global ctx
    del ctx

def test_Mkdir():
    ret = ctx.mkdir(testdir, 0)
    assert(ret == 0)

def test_MkdirFail():
    try:
        ret = ctx.mkdir(testdir)
    except smbc.ExistsError:
        pass
    except:
        assert False
    else:
        assert False
    assert True

def test_ListDir():
    list = ctx.opendir(testdir).getdents()
    assert(len(list) == 2)

def test_Stat():
    st = ctx.stat(testdir)
    mode = st[stat.ST_MODE]
    assert(stat.S_ISDIR(mode))
    assert(stat.S_ISREG(mode) == False)

def test_Rename():
    src = testdir + '/dir1'
    dst = testdir + '/dir2'
    ret = ctx.mkdir(src)
    assert(ret == 0)
    ret = ctx.rename(src, dst)
    assert(ret == 0)

def test_StatFail():
    uri = testdir + '/dir1'
    try:
        ctx.stat(uri)
    except smbc.NoEntryError:
        pass
    except:
        assert False

def test_Rmdir():
    uri = testdir + '/dir2'
    ret = ctx.rmdir(uri)
    assert(ret == 0)

def test_Cleanup():
    ret = ctx.rmdir(testdir)
    assert(ret == 0)




#!/usr/bin/env python
import os
import smbc
import settings
import nose
from nose.plugins.skip import SkipTest


baseurl =  'smb://' + settings.SERVER + "/" + settings.SHARE +"/"


# a nice map from/to smbc constants
smbcType = {
    'WORKGROUP' : smbc.WORKGROUP,
    'SERVER' : smbc.SERVER,
    'FILE_SHARE' : smbc.FILE_SHARE,
    'PRINTER_SHARE' : smbc.PRINTER_SHARE,
    'IPC_SHARE' : smbc.IPC_SHARE,	
	
    smbc.WORKGROUP :  'WORKGROUP',
    smbc.SERVER : 'SERVER',
    smbc.FILE_SHARE : 'FILE_SHARE',
    smbc.PRINTER_SHARE : 'PRINTER_SHARE',
    smbc.IPC_SHARE : 'IPC_SHARE'
}

# another map for system errors TODO can you find them in another module?
EINVAL = 22

def setUp():
    global ctx
    ctx = smbc.Context()
    cb = lambda se, sh, w, u, p: (w, settings.USERNAME, settings.PASSWORD)
    ctx.functionAuthData = cb


def tearDown():
    global ctx
    del ctx

def touch_file(name):
    """
    create a file containing "sample test file" in the test baseurl 
    """
    tmpfile_name = baseurl + name
    dfile = ctx.open(tmpfile_name, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
    dfile.write("sample test file")
    dfile.close
    return tmpfile_name
        
        
def test_xattr_constants():
    assert smbc.XATTR_ACL 
    assert smbc.XATTR_OWNER 
    assert smbc.XATTR_GROUP

    
def test_xattr_get():
    """
    system.nt_sec_desc.<attribute name>
 *                     system.nt_sec_desc.*
 *                     system.nt_sec_desc.*+
 *
 *                  where <attribute name> is one of:
 *
 *                     revision
 *                     owner
 *                     owner+
 *                     group
 *                     group+
 *                     acl:<name or sid>
 *                     acl+:<name or sid
    """
    print("test_xattr")
    furl = touch_file("tmpfile.out")

    # create all combinations of attribute strings
    plus_xattrs = ["%s%s" % (i,j) for i in ["owner", "group", "*"] for j in ["","+"]]
    plus_xattrs.append("revision")
    valid_xatts = ["system.nt_sec_desc." +i for i in  plus_xattrs]

    # check their existence
    for xattr in valid_xatts:
        print("\ttesting %s with %s" % (furl, xattr))
        assert(ctx.getxattr(furl, xattr))
    ctx.open(furl)

def test_xattr_get_error():
    """ Verify that a RuntimeError is raised when passing bad arguments to getxattr()

         Bad arguments include malformed xattrs and unexistent file
    """
    print("test_xattr")
    furl = touch_file("tmpfile.out")

    # create all combinations of attribute strings
    plus_xattrs = ["%s%s" % (i,j) for i in ["owner", "pluto", "*"] for j in ["x","-"]]
    plus_xattrs.append("revisionX")
    invalid_xatts = ["system.nt_sec_desc." +i for i in  plus_xattrs]

    try:
        ctx.getxattr("UNEXISTENT", smbc.XATTR_OWNER)
        assert False, "getxattr should fail with an unexistent file"
    except ValueError as e: 
        (errno,strerror) = e.args 
        assert errno == EINVAL # TODO is it possible to trap an unexistent entity error from smbclient?
        pass

    # check their existence
    for xattr in invalid_xatts:
        print("\ttesting %s with %s" % (furl, xattr))
        try:
            ctx.getxattr(furl, xattr)
            assert False, "getxattr should fail with %s" % xattr
        except ValueError as e:
            (errno,strerror) = e.args 
            assert errno == EINVAL # invalid arguments

    ctx.open(furl)

def test_xattr_set():
    #raise SkipTest("xattr_set to be implemented")
    print("test_xattr_put")
    furl = touch_file("tmpfile_set.out")
    attr_name = smbc.XATTR_ALL
    attrs = ctx.getxattr(furl, attr_name)
    print("attrs(%s): %s" % (attr_name,  attrs))
    ctx.setxattr(furl, attr_name, attrs, smbc.XATTR_FLAG_REPLACE)
    attrs1 = ctx.getxattr(furl, attr_name)
    print("attrs1(%s): %s" % (attr_name,  attrs1))
    assert attrs1 == attrs

@SkipTest
def test_xattr_set_2():
    furl = touch_file("tmpfile_set.out")
    attrs_new = u'REVISION:1,OWNER:RPOLLI\\babel" \
        + ",GROUP:Unix Group\\babel" \
        + ",ACL:RPOLLI\\babel:0/0/0x001e01ff" \
        + ",ACL:Unix Group\\babel:0/0/0x00120089" \
        + ",ACL:Unix Group\\games:0/0/0x001e01ff" \
        + ",ACL:\\Everyone:0/0/0x00120089'
    attr_name = smbc.XATTR_ALL_SID
    attrs_0 = ctx.getxattr(furl, attr_name)
    print("original attrs(%s)" % attrs_0)

    assert attrs_0 != attrs_new, "Old and new attributes are the same:\n%s\n%s\n" % (attrs_0, attrs_new)

    ctx.setxattr(furl, attr_name, attrs_new, smbc.XATTR_FLAG_REPLACE)
    attrs_1 = ctx.getxattr(furl, attr_name)

    print("attrs_1(%s): %s" % (attr_name,  attrs_1))
    assert attrs_1 == attrs_new

def test_xattr_set_error():
    #raise SkipTest("xattr_set to be implemented")
    print("test_xattr_set_error")
    furl = touch_file("tmpfile_set.out")
    attr_name = smbc.XATTR_ALL_SID
    attrs_ok = ctx.getxattr(furl, attr_name)
    attrs = "BAD_VALUE" # causes segfault
    for xa in ["BAD_VALUE", u'REVISION:1,OWNER:RPOLLI\\babel,GROUP:', 0, None]: 
        try:
            ctx.setxattr(furl, attr_name, xa, smbc.XATTR_FLAG_REPLACE)
        except ValueError as e:
            (errno,strerror) = e.args 
            assert errno == EINVAL # invalid arguments
            print("setxattr(%s) raises  %s" % (xa, e))
            pass
        except TypeError as e:
            print("setxattr(%s) raises  %s" % (xa, e))
            pass 
    
    
    
def test_Workgroup():
    l_entries = ctx.opendir('smb://').getdents()
    assert(len(l_entries) > 0)
    for entry in l_entries:
        assert(entry.smbc_type == smbc.WORKGROUP), "Entry %s of type %s, expected %s" % (entry.name, smbcType[entry.smbc_type], smbcType[smbc.WORKGROUP])

def test_Server():
    uri = 'smb://' + settings.WORKGROUP 
    l_entries = ctx.opendir(uri).getdents()
    assert(len(l_entries) > 0)
    for entry in l_entries:
        assert(entry.smbc_type == smbc.SERVER), "Entry %s of type %s, expected %s" % (entry.name, smbcType[entry.smbc_type], smbcType[smbc.SERVER])

def test_Share():
    uri = 'smb://' + settings.SERVER
    l_entries = ctx.opendir(uri).getdents()

    allowed_shares = [smbc.FILE_SHARE, smbc.PRINTER_SHARE, smbc.IPC_SHARE, smbc.COMMS_SHARE]
    assert(len(l_entries) > 0)
    for entry in l_entries:
        assert (entry.smbc_type in allowed_shares), "Entry was %s (%d), expected values: %s" % (smbcType[entry.smbc_type], entry.smbc_type, allowed_shares) 



#!/usr/bin/env python

import smbc
import settings
import sys

def setUp():
    pass

def tearDown():
    pass

def test_AuthSuccess():
    ctx = smbc.Context()
    ctx.optionNoAutoAnonymousLogin = True
    cb = lambda se, sh, w, u, p: (w, settings.USERNAME, settings.PASSWORD)
    ctx.functionAuthData = cb
    uri = 'smb://' + settings.SERVER + '/' + settings.SHARE
    try:
        dir = ctx.opendir(uri)
        print("ok: connection to ", uri)
    except:
        print("fail: connection to ", uri)
        assert False

def test_AuthFailNoauth():
    ctx = smbc.Context()
    ctx.optionNoAutoAnonymousLogin = True
    uri = 'smb://' + settings.SERVER + '/' + settings.SHARE
    try:
        dir = ctx.opendir(uri)
    except smbc.PermissionError:
        print("ok: permission error to ", uri)
        pass
    except:
        print("fail: error connecting to", uri)
        assert False
    else:
        print("fail: error connecting to", uri)
        assert False

def test_AuthFailNopass():
    ctx = smbc.Context()
    ctx.optionNoAutoAnonymousLogin = True
    cb = lambda se, sh, w, u, p: (w, settings.USERNAME, "")
    ctx.functionAuthData = cb
    uri = 'smb://' + settings.SERVER + '/' + settings.SHARE
    try:
        dir = ctx.opendir(uri)
    except smbc.PermissionError:
        pass
    except:
        assert False
    else:
        assert False

def test_AuthFailNoname():
    ctx = smbc.Context()
    ctx.optionNoAutoAnonymousLogin = True
    cb = lambda se, sh, w, u, p: (w, "", "")
    ctx.functionAuthData = cb
    uri = 'smb://' + settings.SERVER + '/' + settings.SHARE
    try:
        dir = ctx.opendir(uri)
    except smbc.PermissionError:
        pass
    except:
        assert False
    else:
        assert False





# This is setting template.
# You must change these parameter if you perform test.
# The following configration will create \\server\share\testdir
# directory, so user should have write permission.

WORKGROUP = "WORKGROUP"
SERVER = "server"
SHARE = "share"
USERNAME = "username"
PASSWORD = "password"
TESTDIR = "testdir"





#!/usr/bin/env python

## Copyright (C) 2002, 2005, 2006, 2007, 2008, 2010, 2011, 2012  Red Hat, Inc
## Copyright (C) 2010  Open Source Solution Technology Corporation
## Authors:
##  Tim Waugh <twaugh@redhat.com>
##  Tsukasa Hamano <hamano@osstech.co.jp>
##  Laurent Coustet <laurent.coustet@clarisys.fr>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""This is a set of Python bindings for the libsmbclient library
from the samba project.

>>> # Directory listing example:
>>> import smbc
>>> ctx = smbc.Context (auth_fn=my_auth_callback_fn)
>>> entries = ctx.opendir ("smb://SERVER").getdents ()
>>> for entry in entries:
...     print entry
<smbc.Dirent object "music" (File share) at 0x7fbd7c42b3a0>
<smbc.Dirent object "IPC$" (IPC share) at 0x7fbd7c42b148>
<smbc.Dirent object "Charlie" (Printer share) at 0x7fbd7c42b3c8>
>>> d = ctx.open ("smb://SERVER/music")

>>> # Write file example:
>>> import smbc
>>> import os
>>> ctx = smbc.Context (auth_fn=my_auth_callback_fn)
>>> file = ctx.open ("smb://SERVER/music/file.txt", os.O_CREAT | os.O_WRONLY)
>>> file.write ("hello")

>>> # Read file example:
>>> import smbc
>>> ctx = smbc.Context (auth_fn=my_auth_callback_fn)
>>> file = ctx.open ("smb://SERVER/music/file.txt")
>>> print file.read()
hello

"""

import subprocess
from setuptools import setup, Extension

def pkgconfig_I(pkg):
    dirs = []
    c = subprocess.Popen(["pkg-config", "--cflags", pkg], stdout=subprocess.PIPE)
    (stdout, stderr) = c.communicate ()
    for p in stdout.decode('ascii').split():
        if p.startswith("-I"):
            dirs.append(p[2:])
    return dirs

setup(
    name="pysmbc",
    version="1.0.16",
    description="Python bindings for libsmbclient",
    long_description=__doc__,
    author=[
        "Tim Waugh <twaugh@redhat.com>",
        "Tsukasa Hamano <hamano@osstech.co.jp>",
        "Roberto Polli <rpolli@babel.it>",
    ],
    url="http://cyberelk.net/tim/software/pysmbc/",
    download_url="http://cyberelk.net/tim/data/pysmbc/",
    license="GPLv2+",
    packages=["smbc"],
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: Unix",
        "Programming Language :: C",
    ],
    ext_modules=[
        Extension("_smbc", [
            "smbc/smbcmodule.c",
            "smbc/context.c",
            "smbc/dir.c",
            "smbc/file.c",
            "smbc/smbcdirent.c"
        ],
        libraries=["smbclient"],
        include_dirs=pkgconfig_I("smbclient")
        )
    ],
)
